﻿

Namespace Charting
    ''' <summary>
    ''' Enhanced Legend control for TeeChart
    ''' </summary>
    Public Class Legend2D
        Implements ILegend
        Public Const MaxLineThickness As Integer = 3

        Private Const MinRows As Integer = 1
        Private Const MinCols As Integer = 1
        Private Const FixedRowHeight As Integer = 20
        Private Const RowMargin As Integer = 0
        Private Const ColMargin As Integer = 0
        Private Const MinColWidth As Integer = 32
        Private Const MaxColumnWidth As Integer = 120
        Private Const FixedViewWidth As Integer = 122
        Protected Const ShadowThickness As Integer = 3
        Private Const HintLabelHeight As Integer = 48
        Private Const HintMargin As Integer = 8
        Private Const ColLabelWidthPercentage As Double = 0.66

        Private m_RowCount As Integer
        Private m_ColumnCount As Integer
        Private m_VisibleRowCount As Integer
        Private m_VisibleColumnCount As Integer

        Private m_RowText As String()
        Private m_RowHeight As Integer()

        Private m_ColumnText As String()
        Private m_ColumnWidth As Integer()

        Private m_RowColor() As Color
        Private m_CellCheckBoxes(,) As CheckBox
        Private m_RowCheckBoxes() As CheckBox
        Private m_ColumnCheckBoxes() As CheckBox
        Private m_EditColumnLabels() As Label
        Private m_EditRowLabels() As Label
        Private m_ViewRowLabels() As Label
        Private m_ViewColumnLabels() As Label
        Private m_EditButton As Button
        Private m_DoneButton As Button

        Private m_HighlightLabel As Label

        Protected m_IsInitialized As Boolean

        Public Sub New()
            InitializeComponent()
            SetStyle(ControlStyles.DoubleBuffer Or ControlStyles.UserPaint Or ControlStyles.AllPaintingInWmPaint, True)
        End Sub

        ''' <summary>
        ''' Initialize the Legend2d to a specifed number of rows and columns
        ''' </summary>
        ''' <param name="rowCaptions">List of captions for rows</param>
        ''' <param name="columnCaptions">List of captions for columns</param>
        ''' <remarks>These zeroth element of the arrays are ignored</remarks>
        Public Sub Initialize(rowCaptions As IList(Of String), columnCaptions As IList(Of String), rowColors As IList(Of Color))
            Dim rowCount As Integer = rowCaptions.Count()
            Dim columnCount As Integer = columnCaptions.Count()

            If rowCount < MinRows Then Throw New ArgumentException("A Legend2D must have at least " & MinRows & " row")
            If columnCount < MinCols Then Throw New ArgumentException("A Legend2D must have at least " & MinCols & "  column")
            If rowCount <> rowColors.Count() Then Throw New ArgumentException("A Legend2D must assign a color to each row")

            m_RowCount = rowCount
            m_ColumnCount = columnCount

            ReDim m_RowText(m_RowCount)
            ReDim m_RowHeight(m_RowCount)
            ReDim m_RowCheckBoxes(m_RowCount)
            ReDim m_EditRowLabels(m_RowCount)
            ReDim m_ViewRowLabels(m_RowCount)
            ReDim m_RowColor(m_RowCount)

            ReDim m_ColumnText(m_ColumnCount)
            ReDim m_ColumnWidth(m_ColumnCount)
            ReDim m_ColumnCheckBoxes(m_ColumnCount)
            ReDim m_EditColumnLabels(m_ColumnCount)
            ReDim m_ViewColumnLabels(m_ColumnCount)
            ReDim m_CellCheckBoxes(m_RowCount, m_ColumnCount)

            For ixRow As Integer = 0 To m_RowCount - 1
                m_RowText(ixRow + 1) = rowCaptions(ixRow)
                m_RowColor(ixRow + 1) = rowColors(ixRow)
            Next

            For ixColumn As Integer = 0 To m_ColumnCount - 1
                m_ColumnText(ixColumn + 1) = columnCaptions(ixColumn)
            Next

            m_IsInitialized = True
        End Sub

        Protected Overridable Sub UpdateChart()

        End Sub

        ''' <summary>
        ''' Refreshes the legend control including dynamic resizing and layout
        ''' </summary>
        Overrides Sub Refresh()
            MyBase.Refresh()
            If Not m_IsInitialized Then Return
            RecalculateWidths()
            InitializeEditPanel()
            InitializeViewPanel()
        End Sub

        Private Sub RecalculateWidths()
            Dim sizer As MeasureString = New MeasureString(20)
            For ixRow As Integer = 0 To m_RowCount
                m_RowHeight(ixRow) = FixedRowHeight
                sizer.Examine(Me, m_RowText(ixRow), Font)
            Next
            m_RowHeight(0) = CInt(FixedRowHeight * 3) 'just use a fixed height

            'Calculate the width of the row labels based on these rules:
            '  Be at least twice as wide as the minimum width of other columns
            '  Don't constrain the columns to be below their minimum width
            '  Don't take too big a percentage of the available width
            '  Don't be wider than necessary to fit all text
            Dim maxAvailableWidth As Integer = 1000
            If Not Me.Parent Is Nothing Then maxAvailableWidth = CInt(Me.Parent.Width * 0.6)
            Dim rowLabelWidth As Integer = Math.Max(MinColWidth * 2,
                                         Math.Min(maxAvailableWidth - (m_ColumnCount * MinColWidth),
                                                  Math.Min(CInt(maxAvailableWidth * ColLabelWidthPercentage), sizer.Width)))

            'Calculate the width of the other columns based on these rules:
            ' Divide the available space evenly among remaining columns
            ' Don't be wider than a maximum defined width
            Dim columnWidth As Integer = Math.Min(MaxColumnWidth, CInt((maxAvailableWidth - rowLabelWidth) / m_ColumnCount))

            m_ColumnWidth(0) = rowLabelWidth
            For ixColumn As Integer = 1 To m_ColumnCount
                m_ColumnWidth(ixColumn) = columnWidth
            Next
        End Sub

        Public Shadows Sub Resize()
            If Not m_IsInitialized Then Return

            Dim isEditing As Boolean = EditPanel.Visible
            Dim isVisible As Boolean = Me.Visible

            Me.Visible = False
            If isEditing Then
                ResizeForEditing(True)
                RecalculateWidths()
                ResizeForEditing(False)
                RefreshEditPanel()
            Else
                ResizeForViewing()
            End If
            Me.Visible = isVisible
        End Sub

        Public ReadOnly Property PreferredViewWidth As Integer
            Get
                Return FixedViewWidth
            End Get
        End Property

        Dim m_PreferredViewHeight As Integer
        Public ReadOnly Property PreferredViewHeight As Integer
            Get
                Return CInt(IIf(m_PreferredViewHeight > 0, m_PreferredViewHeight, Me.Height))
            End Get
        End Property

        Public ReadOnly Property PreferredEditWidth As Integer
            Get
                Return m_ColumnWidth.Sum() + 12
            End Get
        End Property

        Public ReadOnly Property PreferredEditHeight As Integer
            Get
                Return m_RowHeight.Sum() + 16 + HintLabelHeight
            End Get
        End Property

        ''' <summary>
        ''' Returns  helper class for accessing properties of a row
        ''' </summary>
        ''' <param name="rowIndex">1-based index of the row</param>
        Public ReadOnly Property Row(ByVal rowIndex As Integer) As LegendRow
            Get
                Return New LegendRow(Me, rowIndex)
            End Get
        End Property

        ''' <summary>
        ''' Returns  helper class for accessing properties of a column
        ''' </summary>
        ''' <param name="columnIndex">1-based index of the column</param>
        Public ReadOnly Property Column(ByVal columnIndex As Integer) As LegendColumn
            Get
                Return New LegendColumn(Me, columnIndex)
            End Get
        End Property

        ''' <summary>
        ''' Returns  helper class for accessing properties of a cell
        ''' </summary>
        ''' <param name="index">Zero-based index of the cell (for aligning with TeeChart series index</param>
        Public ReadOnly Property Cell(ByVal index As Integer) As LegendCell
            Get
                Dim ixRow As Integer = (index \ m_ColumnCount) + 1
                Dim ixColumn As Integer = (index Mod m_ColumnCount) + 1
                Return New LegendCell(Me, ixRow, ixColumn)
            End Get
        End Property

        ''' <summary>
        ''' Returns the number of columns defined
        ''' </summary>
        Public ReadOnly Property ColumnCount As Integer
            Get
                Return m_ColumnCount
            End Get
        End Property

        ''' <summary>
        ''' Returns the number of rows defined
        ''' </summary>
        Public ReadOnly Property RowCount As Integer
            Get
                Return m_RowCount
            End Get
        End Property

        ''' <summary>
        ''' Returns  helper class for accessing properties of a cell
        ''' </summary>
        ''' <param name="rowIndex">1-based index of the row</param>
        ''' <param name="columnIndex">1-based index of the column</param>
        Public ReadOnly Property Cell(ByVal rowIndex As Integer, ByVal columnIndex As Integer) As LegendCell
            Get
                Return New LegendCell(Me, rowIndex, columnIndex)
            End Get
        End Property

        Private Sub InitializeEditPanel()
            EditPanel.Controls.Clear()

            'Add control to switch back to normal mode
            Dim button As Button = CreateButton(m_ColumnWidth(0) - 16, m_RowHeight(0) - 16, 8, 8, "OK")
            m_DoneButton = button
            AddHandler button.Click, New EventHandler(AddressOf EndEditHandler)
            EditPanel.Controls.Add(button)

            'Assign row labels
            Dim x As Integer = ColMargin
            Dim y As Integer = RowMargin + m_RowHeight(0)
            For ixRow As Integer = 1 To m_RowCount
                Dim label As Label = CreateLabel((m_ColumnWidth(0) - MinColWidth) + 8, m_RowHeight(ixRow), x + 4, y, m_RowText(ixRow), Color.Transparent)
                label.TextAlign = ContentAlignment.MiddleLeft
                label.AutoEllipsis = True
                label.AutoSize = False
                label.Tag = ixRow
                AddHandler label.Click, New EventHandler(AddressOf RowSelectedHandler)
                AddHandler label.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)

                m_EditRowLabels(ixRow) = label
                EditPanel.Controls.Add(label)

                Dim checkBox As CheckBox = CreateCheckBox(MinColWidth, m_RowHeight(ixRow), x + 4 + label.Width, y, "", Color.White, New EventHandler(AddressOf RowClickHandler), ixRow)
                m_RowCheckBoxes(ixRow) = checkBox
                checkBox.CheckAlign = ContentAlignment.MiddleCenter
                checkBox.TextAlign = ContentAlignment.MiddleCenter
                checkBox.AutoEllipsis = True
                AddHandler checkBox.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                EditPanel.Controls.Add(checkBox)
                y += m_RowHeight(ixRow)
            Next

            'Assign column labels
            x = ColMargin + m_ColumnWidth(0)
            y = RowMargin
            'y += m_RowHeight.Sum() / 2 ' Center the control
            For ixColumn As Integer = 1 To m_ColumnCount
                Dim label As Label = CreateLabel(m_ColumnWidth(ixColumn), FixedRowHeight * 2, x, y, m_ColumnText(ixColumn), Color.Transparent)
                label.TextAlign = ContentAlignment.MiddleCenter
                label.AutoEllipsis = True
                label.AutoSize = False
                label.Tag = ixColumn
                AddHandler label.Click, New EventHandler(AddressOf ColumnSelectedHandler)
                AddHandler label.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                EditPanel.Controls.Add(label)
                Dim checkBox As CheckBox = CreateCheckBox(m_ColumnWidth(ixColumn), FixedRowHeight, x, y + FixedRowHeight * 2, "", Color.White, New EventHandler(AddressOf ColumnClickHandler), ixColumn)
                m_EditColumnLabels(ixColumn) = label
                m_ColumnCheckBoxes(ixColumn) = checkBox
                checkBox.CheckAlign = ContentAlignment.MiddleCenter
                checkBox.Text = checkBox.Text
                AddHandler checkBox.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                EditPanel.Controls.Add(checkBox)
                x += m_ColumnWidth(ixColumn)
            Next

            'Assign cells
            y = RowMargin + m_RowHeight(0)
            For ixRow As Integer = 1 To m_RowCount
                x = ColMargin + m_ColumnWidth(0)
                For ixColumn As Integer = 1 To m_ColumnCount
                    Dim checkBox As CheckBox = CreateCheckBox(m_ColumnWidth(ixColumn) - 2, m_RowHeight(ixRow) - 2, x + 1, y + 1, "", m_RowColor(ixRow),
                                                  New EventHandler(AddressOf CellClickHandler), ixRow & ixColumn)
                    checkBox.CheckAlign = ContentAlignment.MiddleCenter
                    m_CellCheckBoxes(ixRow, ixColumn) = checkBox
                    AddHandler checkBox.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                    EditPanel.Controls.Add(checkBox)
                    x += m_ColumnWidth(ixColumn)
                Next
                y += m_RowHeight(ixRow)
            Next

            'Add a hint at the bottom of the display
            Dim hint As Label = CreateLabel(EditPanel.Width - 2 * HintMargin, HintLabelHeight, HintMargin, y + HintMargin,
                                            "", Color.LightGoldenrodYellow)
            hint.Anchor = AnchorStyles.Left Or AnchorStyles.Right or AnchorStyles.Top
            hint.TextAlign = ContentAlignment.MiddleCenter
            hint.Font = New Font(hint.Font.FontFamily, hint.Font.Size, FontStyle.Italic)
            hint.Text = "Click any row or column caption to select that cross-section " _
                      & "or use the checkboxes to select what you wish to display"
            EditPanel.Controls.Add(hint)
        End Sub

        Protected Sub RefreshViewPanel()
            If ViewPanel.Visible Then
                InitializeViewPanel()
            End If
        End Sub

        Private Sub RefreshEditPanel()
            m_DoneButton.Width = m_ColumnWidth(0) - 16
            'Adjust position and size of row labels
            Dim x As Integer = ColMargin
            Dim y As Integer = RowMargin + m_RowHeight(0)
            'y += m_RowHeight.Sum() / 2 ' Center the control
            For ixRow As Integer = 1 To m_RowCount
                With m_EditRowLabels(ixRow)
                    .Width = (m_ColumnWidth(0) - MinColWidth) + 8
                    .Height = m_RowHeight(ixRow)
                    .Left = x + 4
                    .Top = y
                End With
                With m_RowCheckBoxes(ixRow)
                    .Width = MinColWidth
                    .Height = m_RowHeight(ixRow)
                    .Left = x + 4 + m_ColumnWidth(0) - MinColWidth
                    .Top = y
                End With
                y += m_RowHeight(ixRow)
            Next

            'Adjust position and size of column labels
            x = ColMargin + m_ColumnWidth(0)
            y = RowMargin
            For ixColumn As Integer = 1 To m_ColumnCount
                With m_EditColumnLabels(ixColumn)
                    .Width = m_ColumnWidth(ixColumn)
                    .Height = FixedRowHeight * 2
                    .Left = x
                    .Top = y
                End With
                With m_ColumnCheckBoxes(ixColumn)
                    .Width = m_ColumnWidth(ixColumn)
                    .Height = FixedRowHeight
                    .Left = x
                    .Top = y + FixedRowHeight * 2
                End With

                x += m_ColumnWidth(ixColumn)
            Next

            'Adjust position and size of cells
            y = RowMargin + m_RowHeight(0)
            For ixRow As Integer = 1 To m_RowCount
                x = ColMargin + m_ColumnWidth(0)
                For ixColumn As Integer = 1 To m_ColumnCount
                    With m_CellCheckBoxes(ixRow, ixColumn)
                        .Width = m_ColumnWidth(ixColumn) - 2
                        .Height = m_RowHeight(ixRow) - 2
                        .Left = x + 1
                        .Top = y + 1
                    End With

                    x += m_ColumnWidth(ixColumn)
                Next
                y += m_RowHeight(ixRow)
            Next
        End Sub

        Private Sub InitializeViewPanel()
            Dim button As Button
            Dim x As Integer
            Dim y As Integer

            For Each ctrl As Control In ViewPanel.Controls
                RemoveHandler ctrl.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                RemoveHandler ctrl.MouseLeave, New EventHandler(AddressOf ViewPanel_MouseLeave)
                RemoveHandler ctrl.Click, New EventHandler(AddressOf HighlightedItem_Click)
            Next
            ViewPanel.Controls.Clear()

            'Add control to switch back to normal mode
            button = CreateButton(MaxColumnWidth - 8, m_RowHeight(0) - 24, 4, 4, "Edit")
            AddHandler button.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
            AddHandler button.MouseLeave, New EventHandler(AddressOf ViewPanel_MouseLeave)
            AddHandler button.Click, New EventHandler(AddressOf BeginEditHandler)

            m_EditButton = button
            ViewPanel.Controls.Add(button)

            'Assign row labels
            x = 30
            y = RowMargin + m_RowHeight(0) - 16
            m_HighlightLabel = CreateLabel(MaxColumnWidth, FixedRowHeight - 4, 0, y - 2, "", Color.FromArgb(200, 240, 255))
            m_HighlightLabel.Visible = False
            AddHandler m_HighlightLabel.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
            AddHandler m_HighlightLabel.MouseLeave, New EventHandler(AddressOf ViewPanel_MouseLeave)
            AddHandler m_HighlightLabel.Click, New EventHandler(AddressOf HighlightedItem_Click)

            ViewPanel.Controls.Add(m_HighlightLabel)
            m_VisibleRowCount = 0
            For ixRow As Integer = 1 To m_RowCount
                If m_RowCheckBoxes(ixRow).CheckState = CheckState.Unchecked Then Continue For

                m_VisibleRowCount += 1
                Dim colorBar As Label = CreateLabel(16, 8, 8, y + 2, "", m_RowColor(ixRow))
                AddHandler colorBar.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                AddHandler colorBar.MouseLeave, New EventHandler(AddressOf ViewPanel_MouseLeave)
                AddHandler colorBar.Click, New EventHandler(AddressOf HighlightedItem_Click)

                ViewPanel.Controls.Add(colorBar)
                Dim label As Label = CreateLabel(MaxColumnWidth - x - 4, FixedRowHeight - 4, x - 4, y - 2, m_RowText(ixRow), Color.White)
                AddHandler label.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                AddHandler label.MouseLeave, New EventHandler(AddressOf ViewPanel_MouseLeave)
                AddHandler label.Click, New EventHandler(AddressOf HighlightedItem_Click)

                label.Tag = ixRow
                label.TextAlign = ContentAlignment.MiddleLeft
                m_ViewRowLabels(m_VisibleRowCount) = label
                ViewPanel.Controls.Add(label)
                y += m_RowHeight(ixRow)
            Next

            'Assign column labels

            Dim thickness As Integer = 0
            m_VisibleColumnCount = 0
            For ixColumn As Integer = 1 To m_ColumnCount
                If m_ColumnCheckBoxes(ixColumn).CheckState = CheckState.Unchecked Then Continue For

                m_VisibleColumnCount += 1
                thickness += 1
                If thickness > MaxLineThickness Then thickness = 1
                Dim colorBar As Label = CreateLabel(16, thickness, 8, CInt(y + 6 - thickness / 2), "", Color.Gray)
                AddHandler colorBar.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                AddHandler colorBar.MouseLeave, New EventHandler(AddressOf ViewPanel_MouseLeave)
                AddHandler colorBar.Click, New EventHandler(AddressOf HighlightedItem_Click)

                ViewPanel.Controls.Add(colorBar)
                Dim label As Label = CreateLabel(MaxColumnWidth - x - 4, FixedRowHeight - 4, x - 4, y - 2, m_ColumnText(ixColumn), Color.White)
                AddHandler label.MouseEnter, New EventHandler(AddressOf ViewPanel_MouseEnter)
                AddHandler label.MouseLeave, New EventHandler(AddressOf ViewPanel_MouseLeave)
                AddHandler label.Click, New EventHandler(AddressOf HighlightedItem_Click)

                label.Tag = ixColumn
                label.TextAlign = ContentAlignment.MiddleLeft
                label.ForeColor = Color.Gray
                m_ViewColumnLabels(m_VisibleColumnCount) = label
                ViewPanel.Controls.Add(label)
                y += m_RowHeight(1)
            Next
            m_PreferredViewHeight = y
            m_HighlightLabel.SendToBack()

            'This allows for a TeeChartLegend to adjust position based on its CenterLegend property
            ResizeForViewing()
        End Sub

        Private Sub BeginEditHandler(ByVal sender As Object, ByVal e As EventArgs)
            Try


                Me.Visible = False
                ResizeForEditing(True)
                RecalculateWidths()
                ResizeForEditing(False)
                RefreshEditPanel()
                EditPanel.Visible = True
                ViewPanel.Visible = False
                Me.Visible = True
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub EndEditHandler(ByVal sender As Object, ByVal e As EventArgs)
            Try
                Me.Visible = False
                InitializeViewPanel()
                ResizeForViewing()
                EditPanel.Visible = False
                ViewPanel.Visible = True
                Me.Visible = True
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Protected Overridable Sub ResizeForEditing(fullSize As Boolean)
        End Sub

        Protected Overridable Sub ResizeForViewing()
        End Sub

        ' ReSharper disable ParameterHidesMember
        Private Shared Function CreateLabel(width As Integer, height As Integer, left As Integer, top As Integer,
                                     text As String, backColor As Color) As Label
            ' ReSharper restore ParameterHidesMember
            Dim label As New Label
            label.AutoSize = False
            label.AutoEllipsis = True
            label.Width = width
            label.Height = height
            label.Top = top
            label.Left = left
            label.Text = text
            label.BackColor = backColor
            label.UseMnemonic = False ' allows ambersand to be displayed
            Return label
        End Function

        ' ReSharper disable ParameterHidesMember
        Private Shared Function CreateCheckBox(width As Integer, height As Integer, left As Integer, top As Integer,
                                        text As String, backColor As Color,
                                        dlg As EventHandler, tag As Object) As CheckBox
            ' ReSharper restore ParameterHidesMember
            Dim checkBox As New CheckBox
            checkBox.AutoSize = False
            checkBox.Width = width
            checkBox.Height = height
            checkBox.Top = top
            checkBox.Left = left
            checkBox.Text = text
            checkBox.Checked = True
            checkBox.CheckState = CheckState.Checked

            checkBox.BackColor = backColor
            AddHandler checkBox.Click, dlg
            checkBox.Tag = tag
            Return checkBox
        End Function

        ' ReSharper disable ParameterHidesMember
        Private Shared Function CreateButton(width As Integer, height As Integer, left As Integer, top As Integer, text As String) As Button
            ' ReSharper restore ParameterHidesMember
            Dim button As Button = New Button()
            button.Top = top
            button.Left = left
            button.Width = width
            button.Height = height
            button.AutoSize = False
            button.TextAlign = ContentAlignment.MiddleCenter
            button.Text = text
            button.BackColor = Color.FromArgb(240, 240, 240)
            button.FlatStyle = FlatStyle.Flat
            button.FlatAppearance.BorderSize = 0
            Return button
        End Function

        Private Function HasVisibleRows() As Boolean
            For ixRow As Integer = 1 To m_RowCount
                If m_RowCheckBoxes(ixRow).Checked Then Return True
            Next
            Return False
        End Function

        Private Function HasVisibleColumns() As Boolean
            For ixColumn As Integer = 1 To m_ColumnCount
                If m_ColumnCheckBoxes(ixColumn).Checked Then Return True
            Next
            Return False
        End Function

        Private Sub RowSelectedHandler(ByVal sender As Object, ByVal e As EventArgs)
            Try
                Dim control As Label = CType(sender, Label)
                Dim ixRow As Integer = CInt(control.Tag)
                Debug.WriteLine("Selected Column " & ixRow)
                ProcessRowSelected(ixRow)
                UpdateChart()
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub ProcessRowSelected(ByVal selectedRow As Integer)
            For ixColumn As Integer = 1 To m_ColumnCount
                If m_RowCount > 1 Then
                    m_ColumnCheckBoxes(ixColumn).CheckState = CheckState.Indeterminate
                Else
                    m_ColumnCheckBoxes(ixColumn).CheckState = CheckState.Checked
                End If

                For ixRow As Integer = 1 To m_RowCount
                    If ixRow = selectedRow Then
                        m_CellCheckBoxes(ixRow, ixColumn).Checked = True
                    Else
                        m_CellCheckBoxes(ixRow, ixColumn).Checked = False
                    End If
                Next
            Next
            For ixRow As Integer = 1 To m_RowCount
                If ixRow = selectedRow Then
                    m_RowCheckBoxes(ixRow).Checked = True
                Else
                    m_RowCheckBoxes(ixRow).Checked = False
                End If
            Next
        End Sub

        Private Sub ColumnSelectedHandler(ByVal sender As Object, ByVal e As EventArgs)
            Try
                Dim control As Label = CType(sender, Label)
                Dim ixColumn As Integer = CInt(control.Tag)
                Debug.WriteLine("Selected Column " & ixColumn)
                ProcessColumnSelected(ixColumn)
                UpdateChart()
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub ProcessColumnSelected(ByVal selectedColumn As Integer)
            For ixRow As Integer = 1 To m_RowCount
                If m_ColumnCount > 1 Then
                    m_RowCheckBoxes(ixRow).CheckState = CheckState.Indeterminate
                Else
                    m_RowCheckBoxes(ixRow).CheckState = CheckState.Checked
                End If

                For ixColumn As Integer = 1 To m_ColumnCount
                    If ixColumn = selectedColumn Then
                        m_CellCheckBoxes(ixRow, ixColumn).Checked = True
                    Else
                        m_CellCheckBoxes(ixRow, ixColumn).Checked = False
                    End If
                Next
            Next
            For ixColumn As Integer = 1 To m_ColumnCount
                If ixColumn = selectedColumn Then
                    m_ColumnCheckBoxes(ixColumn).Checked = True
                Else
                    m_ColumnCheckBoxes(ixColumn).Checked = False
                End If
            Next
        End Sub

        Private Sub RowClickHandler(sender As Object, e As EventArgs)
            Try
                Dim control As CheckBox = CType(sender, CheckBox)
                Dim ixRow As Integer = CInt(control.Tag)
                Dim isChecked As Boolean = control.Checked
                Debug.WriteLine("Clicked Row " & ixRow & " is " & control.CheckState.ToString())
                ProcessRowClick(ixRow, isChecked)
                UpdateChart()
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub ProcessRowClick(ixRow As Integer, isChecked As Boolean)
            If HasVisibleColumns() Then
                For ixColumn As Integer = 1 To m_ColumnCount
                    If m_ColumnCheckBoxes(ixColumn).Checked Then
                        m_CellCheckBoxes(ixRow, ixColumn).Checked = isChecked
                    End If
                Next
                For ixColumn As Integer = 1 To m_ColumnCount
                    ProcessCellClick(ixRow, ixColumn, m_CellCheckBoxes(ixRow, ixColumn).Checked)
                Next
            ElseIf HasVisibleRows() Then
                Dim columnState As CheckState = CType(IIf(m_RowCount > 1, CheckState.Indeterminate, CheckState.Checked), CheckState)
                For ixColumn As Integer = 1 To m_ColumnCount
                    m_CellCheckBoxes(ixRow, ixColumn).Checked = isChecked
                    If isChecked Then
                        m_ColumnCheckBoxes(ixColumn).CheckState = columnState
                    End If
                Next
            Else
                For ixColumn As Integer = 1 To m_ColumnCount
                    m_ColumnCheckBoxes(ixColumn).Checked = False
                    m_CellCheckBoxes(ixRow, ixColumn).Checked = False
                Next
            End If
        End Sub

        Private Sub ColumnClickHandler(ByVal sender As Object, ByVal e As EventArgs)
            Try
                Dim control As CheckBox = CType(sender, CheckBox)
                Dim ixColumn As Integer = CInt(control.Tag)
                Dim isChecked As Boolean = control.Checked
                Debug.WriteLine("Clicked Column " & ixColumn & " is " & control.CheckState.ToString())
                ProcessColumnClick(ixColumn, isChecked)
                UpdateChart()
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub ProcessColumnClick(ByVal ixColumn As Integer, ByVal isChecked As Boolean)
            If HasVisibleRows() Then
                For ixRow As Integer = 1 To m_RowCount
                    If m_RowCheckBoxes(ixRow).Checked Then
                        m_CellCheckBoxes(ixRow, ixColumn).Checked = isChecked
                    End If
                Next
                For ixRow As Integer = 1 To m_RowCount
                    ProcessCellClick(ixRow, ixColumn, m_CellCheckBoxes(ixRow, ixColumn).Checked)
                Next
            ElseIf HasVisibleColumns() Then
                Dim rowState As CheckState = CType(IIf(m_ColumnCount > 1, CheckState.Indeterminate, CheckState.Checked), CheckState)
                For ixRow As Integer = 1 To m_RowCount
                    m_CellCheckBoxes(ixRow, ixColumn).Checked = isChecked
                    If isChecked Then
                        m_RowCheckBoxes(ixRow).CheckState = CType(rowState, CheckState)
                    End If
                Next
            Else
                For ixRow As Integer = 1 To m_RowCount
                    m_RowCheckBoxes(ixRow).Checked = False
                    m_CellCheckBoxes(ixRow, ixColumn).Checked = False
                Next
            End If
        End Sub

        Private Sub CellClickHandler(ByVal sender As Object, ByVal e As EventArgs)
            Try
                Dim control As CheckBox = CType(sender, CheckBox)
                Dim s As String = CStr(control.Tag)
                Dim ixRow As Integer = CInt(s.Substring(0, 1))
                Dim ixColumn As Integer = CInt(s.Substring(1, 1))
                Dim isChecked As Boolean = control.CheckState <> CheckState.Unchecked
                Debug.WriteLine("Clicked Cell(" & ixRow & "," & ixColumn & ") is " & control.CheckState.ToString())
                ProcessCellClick(ixRow, ixColumn, isChecked)
                UpdateChart()
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub ProcessCellClick(ByVal ixRow As Integer, ByVal ixColumn As Integer, ByVal isChecked As Boolean)
            Dim state As CheckState = CType(IIf(isChecked, CheckState.Checked, CheckState.Unchecked), CheckState)
            Dim rowIndeterminate As Boolean
            For ix As Integer = 1 To m_ColumnCount
                If m_CellCheckBoxes(ixRow, ix).CheckState <> state Then
                    rowIndeterminate = True
                End If
            Next
            Dim colIndeterminate As Boolean
            For ix As Integer = 1 To m_RowCount
                If m_CellCheckBoxes(ix, ixColumn).CheckState <> state Then
                    colIndeterminate = True
                End If
            Next
            If rowIndeterminate Then
                m_RowCheckBoxes(ixRow).CheckState = CheckState.Indeterminate
            Else
                m_RowCheckBoxes(ixRow).CheckState = state
            End If
            If colIndeterminate Then
                m_ColumnCheckBoxes(ixColumn).CheckState = CheckState.Indeterminate
            Else
                m_ColumnCheckBoxes(ixColumn).CheckState = state
            End If
        End Sub

#Region "ILegend Implementation"

        ''' <summary>
        ''' Get or set the label for a row
        ''' </summary>
        ''' <param name="index">1-based index of the logical row</param>
        Private Property RowCaption(index As Integer) As String Implements ILegend.RowCaption
            Get
                If index < 1 OrElse index > m_RowCount Then Throw New ArgumentException("Invalid row index: " & index)

                Return m_RowText(index)
            End Get
            Set(value As String)
                If Not m_IsInitialized Then Return 'avoids exceptions while form is loading
                If index < 1 OrElse index > m_RowCount Then Throw New ArgumentException("Invalid row index: " & index)

                m_RowText(index) = value
                m_RowCheckBoxes(index).Text = value
                m_ViewRowLabels(index).Text = value
            End Set
        End Property

        ''' <summary>
        ''' Get or set the label for a column
        ''' </summary>
        ''' <param name="index">1-based index of the logical column</param>
        Private Property ColumnCaption(index As Integer) As String Implements ILegend.ColumnCaption
            Get
                If index < 1 OrElse index > m_ColumnCount Then Throw New ArgumentException("Invalid column index: " & index)

                Return m_ColumnText(index)
            End Get
            Set(value As String)
                If Not m_IsInitialized Then Return 'avoids exceptions while form is loading
                If index < 1 OrElse index > m_ColumnCount Then Throw New ArgumentException("Invalid column index: " & index)

                m_ColumnText(index) = value
                m_EditColumnLabels(index).Text = value
                m_ViewColumnLabels(index).Text = value
            End Set
        End Property

        ''' <summary>
        ''' Get or set whether or not a row is visible
        ''' </summary>
        ''' <param name="ixRow">1-based index of the logical row</param>
        Protected Property RowVisible(ByVal ixRow As Integer) As Boolean Implements ILegend.RowVisible
            Get
                Return m_RowCheckBoxes(ixRow).Checked
            End Get
            Set(ByVal value As Boolean)
                Dim checkBox As CheckBox = m_RowCheckBoxes(ixRow)
                checkBox.Checked = value
                RowClickHandler(checkBox, EventArgs.Empty)
            End Set
        End Property

        ''' <summary>
        ''' Get or set whether or not a row is visible
        ''' </summary>
        ''' <param name="ixColumn">1-based index of the logical column</param>
        Protected Property ColumnVisible(ByVal ixColumn As Integer) As Boolean Implements ILegend.ColumnVisible
            Get
                Return m_ColumnCheckBoxes(ixColumn).Checked
            End Get
            Set(ByVal value As Boolean)
                Dim checkBox As CheckBox = m_ColumnCheckBoxes(ixColumn)
                checkBox.Checked = value
                ColumnClickHandler(checkBox, EventArgs.Empty)
            End Set
        End Property

        ''' <summary>
        ''' Get or set whether or not the series associated with a cell are visible
        ''' </summary>
        ''' <param name="ixRow">1-based index of the logical row</param>
        ''' <param name="ixColumn">1-based index of the logical column</param>
        Private Property CellVisible(ByVal ixRow As Integer, ByVal ixColumn As Integer) As Boolean Implements ILegend.CellVisible
            Get
                Return m_CellCheckBoxes(ixRow, ixColumn).Checked
            End Get
            Set(ByVal value As Boolean)
                Dim checkBox As CheckBox = m_CellCheckBoxes(ixRow, ixColumn)
                checkBox.Checked = value
                ProcessCellClick(ixRow, ixColumn, value)
            End Set
        End Property

        ''' <summary>
        ''' Get or set whether or not a row is visible
        ''' </summary>
        ''' <param name="ixRow">1-based index of the logical row</param>
        Private Property RowColor(ByVal ixRow As Integer) As Color Implements ILegend.RowColor
            Get
                Return m_RowColor(ixRow)
            End Get
            Set(ByVal value As Color)
                m_RowColor(ixRow) = value
            End Set
        End Property
#End Region

        Private Sub ViewPanel_MouseEnter(ByVal sender As Object, ByVal e As EventArgs) Handles ViewPanel.MouseEnter
            Try
                UpdateHighlight()
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub ViewPanel_MouseLeave(ByVal sender As Object, ByVal e As EventArgs) Handles ViewPanel.MouseLeave
            Try
                UpdateHighlight()
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Dim m_HighlightIndex As Integer = 0
        Private Sub UpdateHighlight()
            If m_VisibleRowCount = 0 OrElse m_VisibleColumnCount = 0 Then Return

            Dim point As Point = ViewPanel.PointToClient(Cursor.Position)
            Dim minY As Integer = m_ViewRowLabels(1).Top
            Dim maxY As Integer = m_ViewColumnLabels(m_VisibleColumnCount).Bottom
            Dim minX As Integer = 0
            Dim maxX As Integer = m_HighlightLabel.Right

            If point.X < minX OrElse point.X >= maxX Then
                SetHighlightIndex(0)
            ElseIf point.Y < minY OrElse point.Y > maxY Then
                SetHighlightIndex(0)
            Else
                SetHighlightIndex(((point.Y - minY) \ FixedRowHeight) + 1)
            End If
        End Sub

        Private Sub SetHighlightIndex(ByVal index As Integer)
            If index = m_HighlightIndex Then Return

            ViewPanel.SuspendDrawing()

            Dim oldLabel As Label = GetLabelFromHighlightIndex(m_HighlightIndex)
            Dim newLabel As Label = GetLabelFromHighlightIndex(index)
            m_HighlightIndex = index

            If oldLabel IsNot Nothing Then
                oldLabel.BackColor = Color.White
            End If
            If newLabel IsNot Nothing Then
                m_HighlightLabel.Top = newLabel.Top
                newLabel.BackColor = m_HighlightLabel.BackColor
                m_HighlightLabel.Visible = True
            Else
                m_HighlightLabel.Visible = False
            End If

            UpdateChart()

            ViewPanel.ResumeDrawing()
        End Sub

        Private Function GetLabelFromHighlightIndex(ByVal index As Integer) As Label
            If index < 0 OrElse index > m_VisibleRowCount + m_VisibleColumnCount Then
                Return Nothing
            ElseIf index <= m_VisibleRowCount Then
                Return m_ViewRowLabels(index)
            Else
                Return m_ViewColumnLabels(index - m_VisibleRowCount)
            End If
        End Function

        Protected ReadOnly Property HighlightedRow As Integer
            Get
                If EditPanel.Visible OrElse
                    m_HighlightIndex <= 0 OrElse m_HighlightIndex > m_VisibleRowCount Then
                    Return 0
                Else
                    Dim label As Label = GetLabelFromHighlightIndex(m_HighlightIndex)
                    If label Is Nothing OrElse label.Tag Is Nothing Then Return 0
                    Return CInt(label.Tag)
                End If
            End Get
        End Property

        Protected ReadOnly Property HighlightedColumn As Integer
            Get
                If EditPanel.Visible OrElse
                    m_HighlightIndex <= m_VisibleRowCount OrElse m_HighlightIndex > m_VisibleRowCount + m_VisibleColumnCount Then
                    Return 0
                Else
                    Dim label As Label = GetLabelFromHighlightIndex(m_HighlightIndex)
                    If label Is Nothing OrElse label.Tag Is Nothing Then Return 0
                    Return CInt(label.Tag)
                End If
            End Get
        End Property

        Private Sub HighlightedItem_Click(ByVal sender As Object, ByVal e As EventArgs)
            Try
                Dim highlightRow As Integer = HighlightedRow
                Dim highlightColumn As Integer = HighlightedColumn
                BeginEditHandler(sender, e)
                If highlightRow > 0 Then
                    ProcessRowSelected(highlightRow)
                    UpdateChart()
                ElseIf highlightColumn > 0 Then
                    ProcessColumnSelected(highlightColumn)
                    UpdateChart()
                End If
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

        Private Sub EditPanel_MouseEnter(ByVal sender As Object, ByVal e As EventArgs) Handles EditPanel.MouseEnter
            Try
                Cursor.Current = Cursors.Arrow
            Catch ex As Exception
                ErrorsFromGUI(, ex)
            End Try
        End Sub

    End Class
End Namespace